from .pandas_context import PandasCSVDataContext
from .sqlalchemy_context import SqlAlchemyDataContext


def get_data_context(context_type, options, *args, **kwargs):
    """Return a data_context object which exposes options to list datasets and get a dataset from
    that context. This is a new API in Great Expectations 0.4, and is subject to rapid change.

    :param context_type: (string) one of "SqlAlchemy" or "PandasCSV"
    :param options: options to be passed to the data context's connect method.
    :return: a new DataContext object
    """
    if context_type == "SqlAlchemy":
        return SqlAlchemyDataContext(options, *args, **kwargs)
    elif context_type == "PandasCSV":
        return PandasCSVDataContext(options, *args, **kwargs)
    else:
        raise ValueError("Unknown data context.")



from __future__ import division

import pandas as pd
import numpy as np
import pytest
from sqlalchemy import create_engine

from great_expectations.dataset import PandasDataset, SqlAlchemyDataset
import great_expectations.dataset.autoinspect as autoinspect


# Taken from the following stackoverflow:
# https://stackoverflow.com/questions/23549419/assert-that-two-dictionaries-are-almost-equal
def assertDeepAlmostEqual(expected, actual, *args, **kwargs):
    """
    Assert that two complex structures have almost equal contents.

    Compares lists, dicts and tuples recursively. Checks numeric values
    using pyteset.approx and checks all other values with an assertion equality statement
    Accepts additional positional and keyword arguments and pass those
    intact to pytest.approx() (that's how you specify comparison
    precision).

    """
    is_root = '__trace' not in kwargs
    trace = kwargs.pop('__trace', 'ROOT')
    try:
        # if isinstance(expected, (int, float, long, complex)):
        if isinstance(expected, (int, float, complex)):
            assert expected == pytest.approx(actual, *args, **kwargs)
        elif isinstance(expected, (list, tuple, np.ndarray)):
            assert len(expected) == len(actual)
            for index in range(len(expected)):
                v1, v2 = expected[index], actual[index]
                assertDeepAlmostEqual(v1, v2,
                                      __trace=repr(index), *args, **kwargs)
        elif isinstance(expected, dict):
            assert set(expected) == set(actual)
            for key in expected:
                assertDeepAlmostEqual(expected[key], actual[key],
                                      __trace=repr(key), *args, **kwargs)
        else:
            assert expected == actual
    except AssertionError as exc:
        exc.__dict__.setdefault('traces', []).append(trace)
        if is_root:
            trace = ' -> '.join(reversed(exc.traces))
            exc = AssertionError("%s\nTRACE: %s" % (str(exc), trace))
        raise exc


def get_dataset(dataset_type, data, autoinspect_func=autoinspect.columns_exist):
    """For Pandas, data should be either a DataFrame or a dictionary that can
    be instantiated as a DataFrame.
    For SQL, data should have the following shape:
        {
            'table':
                'table': SqlAlchemy Table object
                named_column: [list of values]
        }

    """
    if dataset_type == 'PandasDataset':
        return PandasDataset(data, autoinspect_func=autoinspect_func)
    elif dataset_type == 'SqlAlchemyDataset':
        # Create a new database

        engine = create_engine('sqlite://')

        # Add the data to the database as a new table
        df = pd.DataFrame(data)
        df.to_sql(name='test_data', con=engine, index=False)

        # Build a SqlAlchemyDataset using that database
        return SqlAlchemyDataset('test_data', engine=engine, autoinspect_func=autoinspect_func)
    else:
        raise ValueError("Unknown dataset_type " + str(dataset_type))


def candidate_test_is_on_temporary_notimplemented_list(context, expectation_type):
    if context == "SqlAlchemyDataset":
        return expectation_type in [
            #"expect_column_to_exist",
            #"expect_table_row_count_to_be_between",
            #"expect_table_row_count_to_equal",
            #"expect_table_columns_to_match_ordered_list",
            #"expect_column_values_to_be_unique",
            # "expect_column_values_to_not_be_null",
            # "expect_column_values_to_be_null",
            "expect_column_values_to_be_of_type",
            "expect_column_values_to_be_in_type_list",
            # "expect_column_values_to_be_in_set",
            # "expect_column_values_to_not_be_in_set",
            # "expect_column_values_to_be_between",
            "expect_column_values_to_be_increasing",
            "expect_column_values_to_be_decreasing",
            # "expect_column_value_lengths_to_be_between",
            # "expect_column_value_lengths_to_equal",
            "expect_column_values_to_match_regex",
            "expect_column_values_to_not_match_regex",
            "expect_column_values_to_match_regex_list",
            "expect_column_values_to_not_match_regex_list",
            "expect_column_values_to_match_strftime_format",
            "expect_column_values_to_be_dateutil_parseable",
            "expect_column_values_to_be_json_parseable",
            "expect_column_values_to_match_json_schema",
            #"expect_column_mean_to_be_between",
            #"expect_column_median_to_be_between",
            "expect_column_stdev_to_be_between",
            #"expect_column_unique_value_count_to_be_between",
            #"expect_column_proportion_of_unique_values_to_be_between",
            "expect_column_most_common_value_to_be_in_set",
            #"expect_column_sum_to_be_between",
            #"expect_column_min_to_be_between",
            #"expect_column_max_to_be_between",
            "expect_column_chisquare_test_p_value_to_be_greater_than",
            "expect_column_bootstrapped_ks_test_p_value_to_be_greater_than",
            "expect_column_kl_divergence_to_be_less_than",
            "expect_column_parameterized_distribution_ks_test_p_value_to_be_greater_than",
            "expect_column_pair_values_to_be_equal",
            "expect_column_pair_values_A_to_be_greater_than_B",
            "expect_column_pair_values_to_be_in_set",
        ]
    return False


def evaluate_json_test(dataset, expectation_type, test):
    """
    This method will evaluate the result of a test build using the Great Expectations json test format.

    NOTE: Tests can be suppressed for certain data types if the test contains the Key 'suppress_test_for' with a list
        of Dataset types to suppress, such as ['SQLAlchemy', 'Pandas'].

    :param dataset: (Dataset) A great expectations Dataset
    :param expectation_type: (string) the name of the expectation to be run using the test input
    :param test: (dict) a dictionary containing information for the test to be run. The dictionary must include:
        - title: (string) the name of the test
        - exact_match_out: (boolean) If true, match the 'out' dictionary exactly against the result of the expectation
        - in: (dict or list) a dictionary of keyword arguments to use to evaluate the expectation or a list of positional arguments
        - out: (dict) the dictionary keys against which to make assertions. Unless exact_match_out is true, keys must\
            come from the following list:
              - success
              - observed_value
              - unexpected_index_list
              - unexpected_list
              - details
              - traceback_substring (if present, the string value will be expected as a substring of the exception_traceback)
    :return: None. asserts correctness of results.
    """

    dataset.set_default_expectation_argument('result_format', 'COMPLETE')

    if 'title' not in test:
        raise ValueError(
            "Invalid test configuration detected: 'title' is required.")

    if 'exact_match_out' not in test:
        raise ValueError(
            "Invalid test configuration detected: 'exact_match_out' is required.")

    if 'in' not in test:
        raise ValueError(
            "Invalid test configuration detected: 'in' is required.")

    if 'out' not in test:
        raise ValueError(
            "Invalid test configuration detected: 'out' is required.")

    # Pass the test if we are in a test condition that is a known exception

    # Known condition: SqlAlchemy does not support parse_strings_as_datetimes
    if 'parse_strings_as_datetimes' in test['in'] and isinstance(dataset, SqlAlchemyDataset):
        return

    # Known condition: SqlAlchemy does not support allow_cross_type_comparisons
    if 'allow_cross_type_comparisons' in test['in'] and isinstance(dataset, SqlAlchemyDataset):
        return

    try:
        # Support tests with positional arguments
        if isinstance(test['in'], list):
            result = getattr(dataset, expectation_type)(*test['in'])
        # As well as keyword arguments
        else:
            result = getattr(dataset, expectation_type)(**test['in'])

    except NotImplementedError:
        # Note: This method of checking does not look for false negatives: tests that are incorrectly on the notimplemented_list
        assert candidate_test_is_on_temporary_notimplemented_list(
            dataset.__class__.__name__, expectation_type), "Error: this test was supposed to return NotImplementedError"
        return

    if 'suppress_test_for' in test:
        # Optionally suppress the test for specified Dataset types
        if 'SQLAlchemy' in test['suppress_test_for'] and isinstance(dataset, SqlAlchemyDataset):
            return
        if 'Pandas' in test['suppress_test_for'] and isinstance(dataset, PandasDataset):
            return

    # Check results
    if test['exact_match_out'] is True:
        assert test['out'] == result

    else:
        for key, value in test['out'].items():
            # Apply our great expectations-specific test logic

            if key == 'success':
                assert result['success'] == value

            elif key == 'observed_value':
                # assert np.allclose(result['result']['observed_value'], value)
                assert value == result['result']['observed_value']

            elif key == 'unexpected_index_list':
                if isinstance(dataset, SqlAlchemyDataset):
                    pass
                else:
                    assert result['result']['unexpected_index_list'] == value

            elif key == 'unexpected_list':
                assert result['result']['unexpected_list'] == value, "expected " + \
                    str(value) + " but got " + \
                    str(result['result']['unexpected_list'])

            elif key == 'details':
                assert result['result']['details'] == value

            elif key == 'traceback_substring':
                assert result['exception_info']['raised_exception']
                assert value in result['exception_info']['exception_traceback'], "expected to find " + \
                    value + " in " + \
                    result['exception_info']['exception_traceback']

            else:
                raise ValueError(
                    "Invalid test specification: unknown key " + key + " in 'out'")



###
###
#
# This file should not be modified. To adjust test cases, edit the related json file(s).
#
###
###


import os
import json
import glob
import warnings

from tests.test_utils import get_dataset, candidate_test_is_on_temporary_notimplemented_list, evaluate_json_test

contexts = ['PandasDataset', 'SqlAlchemyDataset']


def pytest_generate_tests(metafunc):

    # Load all the JSON files in the directory
    dir_path = os.path.dirname(os.path.realpath(__file__))
    test_configuration_files = glob.glob(dir_path+'/*.json')

    parametrized_tests = []
    ids = []
    for c in contexts:
        for filename in test_configuration_files:
            file = open(filename)
            test_configuration = json.load(file)

            if candidate_test_is_on_temporary_notimplemented_list(c, test_configuration["expectation_type"]):
                warnings.warn("Skipping generation of tests for expectation " + test_configuration["expectation_type"] +
                              " and context " + c)
            else:
                for d in test_configuration['datasets']:
                    my_dataset = get_dataset(c, d["data"])

                    for test in d["tests"]:
                        parametrized_tests.append({
                            "expectation_type": test_configuration["expectation_type"],
                            "dataset": my_dataset,
                            "test": test,
                        })

                        ids.append(
                            c+":"+test_configuration["expectation_type"]+":"+test["title"])

    metafunc.parametrize(
        "test_case",
        parametrized_tests,
        ids=ids
    )


def test_case_runner(test_case):
    # Note: this should never be done in practice, but we are wiping expectations to reuse datasets during testing.
    test_case["dataset"]._initialize_expectations()

    evaluate_json_test(
        test_case["dataset"],
        test_case["expectation_type"],
        test_case["test"]
    )



"""
Autoinspect utilities to automatically generate expectations by evaluating a dataset.
"""
from __future__ import division

import warnings
from six import string_types

from .util import create_multiple_expectations


class AutoInspectError(Exception):
    """Exception raised for errors in autoinspection.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


def columns_exist(inspect_dataset):
    """
    This function will take a dataset and add expectations that each column present exists.

    Args:
        inspect_dataset (great_expectations.dataset): The dataset to inspect and to which to add expectations.
    """

    # Attempt to get column names. For pandas, columns is just a list of strings
    if not hasattr(inspect_dataset, "columns"):
        warnings.warn(
            "No columns list found in dataset; no autoinspection performed.")
        return
    elif isinstance(inspect_dataset.columns[0], string_types):
        columns = inspect_dataset.columns
    elif isinstance(inspect_dataset.columns[0], dict) and "name" in inspect_dataset.columns[0]:
        columns = [col['name'] for col in inspect_dataset.columns]
    else:
        raise AutoInspectError(
            "Unable to determine column names for this dataset.")

    create_multiple_expectations(
        inspect_dataset, columns, "expect_column_to_exist")



import unittest

import great_expectations as ge


class TestUtilMethods(unittest.TestCase):
    def test_expect_file_hash_to_equal(self):
        # Test for non-existent file
        try:
            ge.expect_file_hash_to_equal('abc', value='abc')
        except IOError:
            pass
        # Test for non-existent hash algorithm
        test_file = './tests/test_sets/Titanic.csv'
        try:
            ge.expect_file_hash_to_equal(test_file,
                                         hash_alg='md51',
                                         value='abc')
        except ValueError:
            pass
        # Test non-matching hash value
        self.assertFalse(ge.expect_file_hash_to_equal(test_file,
                                                      value='abc'))
        # Test matching hash value with default algorithm
        self.assertTrue(ge.expect_file_hash_to_equal(test_file,
                                                     value='63188432302f3a6e8c9e9c500ff27c8a'))
        # Test matching hash value with specified algorithm
        self.assertTrue(ge.expect_file_hash_to_equal(test_file,
                                                     value='f89f46423b017a1fc6a4059d81bddb3ff64891e3c81250fafad6f3b3113ecc9b',
                                                     hash_alg='sha256'))

    def test_expect_file_size_to_be_between(self):
        # Test for non-existent file
        try:
            ge.expect_file_size_to_be_between('abc', 0, 10000)
        except OSError:
            pass
        test_file = './tests/test_sets/Titanic.csv'
        # Test minsize not an integer
        try:
            ge.expect_file_size_to_be_between(test_file, '0', 10000)
        except TypeError:
            pass
        # Test maxsize not an integer
        try:
            ge.expect_file_size_to_be_between(test_file, 0, '10000')
        except TypeError:
            pass
        # Test minsize less than 0
        try:
            ge.expect_file_size_to_be_between(test_file, -1, 10000)
        except ValueError:
            pass
        # Test maxsize less than 0
        try:
            ge.expect_file_size_to_be_between(test_file, 0, -1)
        except ValueError:
            pass
        # Test minsize > maxsize
        try:
            ge.expect_file_size_to_be_between(test_file, 10000, 0)
        except ValueError:
            pass
        # Test file size not in range
        self.assertFalse(
            ge.expect_file_size_to_be_between(test_file, 0, 10000))
        # Test file size in range
        self.assertTrue(ge.expect_file_size_to_be_between(
            test_file, 70000, 71000))

    def test_expect_file_to_exist(self):
        # Test for non-existent file
        self.assertFalse(ge.expect_file_to_exist('abc'))
        # Test for existing file
        self.assertTrue(ge.expect_file_to_exist(
            './tests/test_sets/Titanic.csv'))

    def test_expect_file_unique_column_names_csv(self):
        # Test for non-existent file
        try:
            ge.expect_file_unique_column_names_csv('abc')
        except IOError:
            pass
        # Test for non-unique column names
        test_file = './tests/test_sets/same_column_names.csv'
        self.assertFalse(ge.expect_file_unique_column_names_csv(test_file,
                                                                sep='|',
                                                                skipLines=2))
        # Test for unique column names
        test_file = './tests/test_sets/Titanic.csv'
        self.assertTrue(ge.expect_file_unique_column_names_csv(test_file))

    def test_expect_file_valid_json(self):
        # Test for non-existent file
        try:
            ge.expect_file_valid_json('abc')
        except IOError:
            pass

        # Test invalid JSON file
        test_file = './tests/test_sets/invalid_json_file.json'
        self.assertFalse(ge.expect_file_valid_json(test_file))

        # Test valid JSON file
        test_file = './tests/test_sets/titanic_expectations.json'
        self.assertTrue(ge.expect_file_valid_json(test_file))

        # Test valid JSON file with non-matching schema
        test_file = './tests/test_sets/json_test1_against_schema.json'
        schema_file = './tests/test_sets/sample_schema.json'
        self.assertFalse(ge.expect_file_valid_json(
            test_file, schema=schema_file))

        # Test valid JSON file with valid schema
        test_file = './tests/test_sets/json_test2_against_schema.json'
        schema_file = './tests/test_sets/sample_schema.json'
        self.assertTrue(ge.expect_file_valid_json(
            test_file, schema=schema_file))



# This code is early and experimental. eventually it will live in a RawFileDataset class, per issue 321. In the
# meantime, please use, enjoy, and share feedback.

import csv
import hashlib
import json
import jsonschema
import os.path


def expect_file_hash_to_equal(filename, value, hash_alg='md5'):
    """
    Return True or False indicating whether the hash matches the specified value for the default (md5) or user-specified hash algorithm

    Parameters
    ----------
    filename : string
        file on which the hash is computed
    value : string
        value to compare to computed hash
    hash_alg : string, default='md5'
        hash alogorithm to use. See hashlib.algorithms_available for supported algorithms.

    Returns
    -------
    True if the computed hash matches the specified value; False otherwise

    Raises
    ------
    IOError
        if there is a problem reading the specified file
    ValueError
        if the specified hash algorithm is not defined by hashlib

    """
    success = False
    try:
        hash = hashlib.new(hash_alg)
        # Limit file reads to 64 KB chunks at a time
        BLOCKSIZE = 65536
        try:
            with open(filename, 'rb') as file:
                file_buffer = file.read(BLOCKSIZE)
                while len(file_buffer) > 0:
                    hash.update(file_buffer)
                    file_buffer = file.read(BLOCKSIZE)
                success = hash.hexdigest() == value
        except IOError:
            raise
    except ValueError:
        raise
    return success


def expect_file_size_to_be_between(filename, minsize, maxsize):
    """
    Return True or False indicating whether the file size (in bytes) is (inclusively) between two values.

    Parameters
    ----------
    filename : string
        file to check file size
    minsize : integer
        minimum file size
    maxsize : integer
        maximum file size

    Returns
    -------
    True if the computed hash matches the specified value; False otherwise

    Raises
    ------
    OSError
        if there is a problem reading the specified file
    TypeError
        if minsize or maxsize are not integers
    ValueError
        if there is a problem with the integer value of minsize or maxsize

    """
    try:
        size = os.path.getsize(filename)
    except OSError:
        raise
    if type(minsize) != int:
        raise TypeError('minsize must be an integer')
    if type(maxsize) != int:
        raise TypeError('maxsize must be an integer')
    if minsize < 0:
        raise ValueError('minsize must be greater than of equal to 0')
    if maxsize < 0:
        raise ValueError('maxsize must be greater than of equal to 0')
    if minsize > maxsize:
        raise ValueError('maxsize must be greater than of equal to minsize')
    if (size >= minsize) and (size <= maxsize):
        return True
    else:
        return False


def expect_file_to_exist(filename):
    """
    Return True or False indicating whether the specified file exists

    Parameters
    ----------
    filename : string
        file to check for existence

    Returns
    -------
    True if the specified file exists; False otherwise

    """
    if os.path.isfile(filename):
        return True
    else:
        return False


def expect_file_unique_column_names_csv(filename,
                                        skipLines=0,
                                        sep=',',
                                        quoteChar='"',
                                        quot=csv.QUOTE_MINIMAL,
                                        doubleQuote=True,
                                        skipInitialSpace=False,
                                        escapeChar=None):
    """
    Return True or False indicating whether the specified CSV file has unique column names.

    Parameters
    ----------
    filename : string
        file on which the hash is computed
    skipLines : integer
        number of rows or lines to skip before reading the header line
    sep : string
        delimiter used for parsing CSV lines
    quoteChar : string
        one-character string used to quote fields containing special characters
    quot :
        controls when quotes should be recognised by the reader
    doubleQuote : boolean
        controls how instances of quotechar appearing inside a field should themselves be quoted
    skipInitialSpace : boolean
        when True, whitespace immediately following the delimiter is ignored
    escapeChar : string
        one-char string which removes any special meaning from the following character


    Returns
    -------
    True if the column names are unique; False otherwise

    Raises
    ------
    IOError
        if there is a problem reading the specified file
    csv.Error
        if there is an error in the csv methods

    """
    success = False
    try:
        with open(filename, 'r') as f:
            reader = csv.reader(f, delimiter=sep, quotechar=quoteChar, quoting=quot, doublequote=doubleQuote,
                                skipinitialspace=skipInitialSpace, escapechar=escapeChar, strict=True)
            if skipLines > 0:
                for i in range(0, skipLines):
                    next(reader)
            colnames = next(reader)
            if len(set(colnames)) == len(colnames):
                success = True
    except IOError:
        raise
    except csv.Error:
        raise
    return success


def expect_file_valid_json(filename, schema=None):
    """
    Return True or False indicating whether the specified file is valid JSON.

    Parameters
    ----------
    filename : string
        file to test as valid JSON
    schema : string
        optional JSON schema file on which JSON data file is validated against


    Returns
    -------
    True if the file is valid JSON; False otherwise

    Raises
    ------
    Error
        if there is a problem reading a specified file or parsing as JSON
    jsonschema.SchemaError
        if there is an error in the provided schema

    """
    success = False
    if schema is None:
        try:
            with open(filename, 'r') as f:
                json.load(f)
            success = True
        except ValueError:
            success = False
    else:
        try:
            with open(schema, 'r') as s:
                schema_data = s.read()
            sdata = json.loads(schema_data)
            with open(filename, 'r') as f:
                json_data = f.read()
            jdata = json.loads(json_data)
            jsonschema.validate(jdata, sdata)
            success = True
        except jsonschema.ValidationError:
            success = False
        except jsonschema.SchemaError:
            raise
        except:
            raise
    return success



import unittest
import json
import numpy as np
import sys

import great_expectations as ge
sys.path.append("./tests")
from test_utils import assertDeepAlmostEqual
sys.path.append("./great_expectations/dataset")
from util import is_valid_continuous_partition_object


class TestDistributionalExpectations(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestDistributionalExpectations, self).__init__(*args, **kwargs)
        self.D = ge.read_csv(
            './tests/test_sets/distributional_expectations_data_test.csv')

        with open('./tests/test_sets/test_partitions.json', 'r') as infile:
            self.test_partitions = json.loads(infile.read())

    def test_expect_column_chisquare_test_p_value_to_be_greater_than(self):
        T = [
            {
                'args': ['categorical_fixed'],
                'kwargs': {
                    'partition_object': self.test_partitions['categorical_fixed'],
                    'p': 0.05
                },
                'out': {'success': True, 'observed_value': 1.}
            },
            {
                'args': ['categorical_fixed'],
                'kwargs': {
                    'partition_object': self.test_partitions['categorical_fixed_alternate'],
                    'p': 0.05
                },
                'out': {'success': False, 'observed_value': 5.1397782097623862e-53}
            },
            {
                'args': ['categorical_fixed'],
                'kwargs': {
                    'partition_object': self.test_partitions['categorical_fixed_alternate'],
                    'p': 0.05, 'result_format': 'SUMMARY'
                },
                'out': {'success': False, 'observed_value': 5.1397782097623862e-53,
                        'details': {
                            'observed_partition': {
                                'values': [u'A', u'B', u'C'],
                                'weights': [540, 320, 140]
                            },
                            'expected_partition': {
                                'values': [u'A', u'B', u'C'],
                                'weights': [333.3333333333333, 333.3333333333333, 333.3333333333333]
                            }
                        }
                        }
            }
        ]
        for t in T:
            out = self.D.expect_column_chisquare_test_p_value_to_be_greater_than(
                *t['args'], **t['kwargs'])
            self.assertEqual(t['out']['success'], out['success'])
            self.assertEqual(t['out']['observed_value'],
                             out['result']['observed_value'])
            if 'result_format' in t['kwargs'] and t['kwargs']['result_format'] == 'SUMMARY':
                self.assertDictEqual(
                    t['out']['details'], out['result']['details'])

    def test_expect_column_chisquare_test_p_value_to_be_greater_than_new_categorical_val(self):
        # Note: Chisquare test with true zero expected could be treated subtly. Here, we tolerate a warning from stats.
        categorical_list = (['A'] * 25) + (['B'] * 25) + \
            (['C'] * 25) + (['D'] * 25)
        df = ge.dataset.PandasDataset({'categorical': categorical_list})

        out = df.expect_column_chisquare_test_p_value_to_be_greater_than(
            'categorical', self.test_partitions['categorical_fixed_alternate'])
        self.assertEqual(out['success'], False)

        out = df.expect_column_chisquare_test_p_value_to_be_greater_than(
            'categorical', self.test_partitions['categorical_fixed_alternate'], tail_weight_holdout=0.25)
        self.assertEqual(out['success'], True)

    def test_expect_column_chisquare_test_p_value_to_be_greater_than_missing_categorical_val(self):
        categorical_list = (['A'] * 61) + (['B'] * 39)
        df = ge.dataset.PandasDataset({'categorical': categorical_list})
        out = df.expect_column_chisquare_test_p_value_to_be_greater_than(
            'categorical', self.test_partitions['categorical_fixed'])
        self.assertEqual(out['success'], False)

    def test_expect_column_kl_divergence_to_be_less_than_discrete(self):
        T = [
            {
                'args': ['categorical_fixed'],
                'kwargs': {
                    'partition_object': self.test_partitions['categorical_fixed'],
                    'threshold': 0.1
                },
                'out': {'success': True, 'observed_value': 0.}
            },
            {
                'args': ['categorical_fixed'],
                'kwargs': {
                    'partition_object': self.test_partitions['categorical_fixed_alternate'],
                    'threshold': 0.1
                },
                'out': {'success': False, 'observed_value': 0.12599700286677529}
            },
            {
                'args': ['categorical_fixed'],
                'kwargs': {
                    'partition_object': self.test_partitions['categorical_fixed_alternate'],
                    'threshold': 0.1, 'result_format': 'SUMMARY'
                },
                'out': {'success': False, 'observed_value': 0.12599700286677529,
                        'details': {
                            'observed_partition': {
                                'weights': [0.54, 0.32, 0.14],
                                'values': [u'A', u'B', u'C']},
                            'expected_partition': {
                                'weights': [0.3333333333333333, 0.3333333333333333, 0.3333333333333333],
                                'values': [u'A', u'B', u'C']
                            }
                        }
                        }
            }
        ]
        for t in T:
            out = self.D.expect_column_kl_divergence_to_be_less_than(
                *t['args'], **t['kwargs'])
            self.assertTrue(np.allclose(out['success'], t['out']['success']))
            self.assertTrue(np.allclose(
                out['result']['observed_value'], t['out']['observed_value']))
            if 'result_format' in t['kwargs'] and t['kwargs']['result_format'] == 'SUMMARY':
                self.assertDictEqual(
                    out['result']['details'], t['out']['details'])

    def test_expect_column_kl_divergence_to_be_less_than_discrete_holdout(self):
        df = ge.dataset.PandasDataset({'a': ['a', 'a', 'b', 'c']})
        out = df.expect_column_kl_divergence_to_be_less_than('a',
                                                             {'values': ['a', 'b'], 'weights': [
                                                                 0.6, 0.4]},
                                                             threshold=0.1,
                                                             tail_weight_holdout=0.1)
        self.assertEqual(out['success'], True)
        self.assertTrue(np.allclose(
            out['result']['observed_value'], [0.099431384003497381]))

        out = df.expect_column_kl_divergence_to_be_less_than('a',
                                                             {'values': ['a', 'b'], 'weights': [
                                                                 0.6, 0.4]},
                                                             threshold=0.1,
                                                             tail_weight_holdout=0.05)
        self.assertEqual(out['success'], False)
        self.assertTrue(np.isclose(
            out['result']['observed_value'], [0.23216776319077681]))

        out = df.expect_column_kl_divergence_to_be_less_than('a',
                                                             {'values': ['a', 'b'], 'weights': [
                                                                 0.6, 0.4]},
                                                             threshold=0.1)
        self.assertEqual(out['success'], False)
        self.assertEqual(out['result']['observed_value'], None)

    def test_expect_column_bootstrapped_ks_test_p_value_to_be_greater_than(self):
        T = [
            {
                'args': ['norm_0_1'],
                'kwargs': {'partition_object': self.test_partitions['norm_0_1_auto'], "p": 0.05},
                'out': {'success': True, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['norm_0_1'],
                'kwargs':{'partition_object': self.test_partitions['norm_0_1_uniform'], "p": 0.05},
                'out':{'success': True, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['norm_0_1'],
                'kwargs':{'partition_object': self.test_partitions['norm_0_1_ntile'], "p": 0.05},
                'out':{'success': True, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['norm_0_1'],
                'kwargs':{'partition_object': self.test_partitions['norm_0_1_kde'], "p": 0.05},
                'out':{'success': True, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['norm_1_1'],
                'kwargs':{'partition_object': self.test_partitions['norm_0_1_auto'], "p": 0.05},
                'out':{'success': False, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['norm_1_1'],
                'kwargs':{'partition_object': self.test_partitions['norm_0_1_uniform'], "p": 0.05},
                'out':{'success': False, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['norm_1_1'],
                'kwargs':{'partition_object': self.test_partitions['norm_0_1_ntile'], "p": 0.05},
                'out':{'success': False, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['norm_1_1'],
                'kwargs':{'partition_object': self.test_partitions['norm_0_1_kde'], "p": 0.05},
                'out':{'success': False, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['bimodal'],
                'kwargs':{'partition_object': self.test_partitions['bimodal_auto'], "p": 0.05},
                'out':{'success': True, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['bimodal'],
                'kwargs':{'partition_object': self.test_partitions['bimodal_kde'], "p": 0.05},
                'out':{'success': True, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['bimodal'],
                'kwargs':{'partition_object': self.test_partitions['norm_0_1_auto'], "p": 0.05,
                          'include_config': True},
                'out':{'success': False, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['bimodal'],
                'kwargs':{'partition_object': self.test_partitions['norm_0_1_uniform'], "p": 0.05},
                'out':{'success': False, 'observed_value': "RANDOMIZED"}
            },
            {
                'args': ['bimodal'],
                'kwargs': {'partition_object': self.test_partitions['norm_0_1_uniform'], "p": 0.05, 'result_format': 'SUMMARY'},
                'out': {'success': False, 'observed_value': "RANDOMIZED",
                        'details': {
                            'expected_cdf': {
                                'cdf_values': [0.0, 0.001, 0.009000000000000001, 0.056, 0.184, 0.429, 0.6779999999999999, 0.8899999999999999, 0.9689999999999999, 0.9929999999999999, 0.9999999999999999],
                                'x': [-3.721835843971108, -3.02304158492966, -2.324247325888213, -1.625453066846767, -0.926658807805319, -0.227864548763872, 0.470929710277574, 1.169723969319022, 1.868518228360469, 2.567312487401916, 3.266106746443364]
                            },
                            'observed_partition': {
                                'weights': [0.001, 0.006, 0.022, 0.07, 0.107, 0.146, 0.098, 0.04, 0.01, 0.0, 0.5],
                                'bins': [-3.721835843971108, -3.02304158492966, -2.324247325888213, -1.625453066846767, -0.926658807805319, -0.227864548763872, 0.470929710277574, 1.169723969319022, 1.868518228360469, 2.567312487401916, 3.266106746443364, 12.8787297644972]
                            },
                            'bootstrap_samples': 1000,
                            'observed_cdf': {
                                'cdf_values': [0, 0.001, 0.007, 0.028999999999999998, 0.099, 0.20600000000000002, 0.352, 0.44999999999999996, 0.48999999999999994, 0.49999999999999994, 0.49999999999999994, 1.0],
                                'x': [-3.721835843971108, -3.02304158492966, -2.324247325888213, -1.625453066846767, -0.926658807805319, -0.227864548763872, 0.470929710277574, 1.169723969319022, 1.868518228360469, 2.567312487401916, 3.266106746443364, 12.8787297644972]
                            },
                            'expected_partition': {
                                'weights': [0.001, 0.008, 0.047, 0.128, 0.245, 0.249, 0.212, 0.079, 0.024, 0.007],
                                'bins': [-3.721835843971108, -3.02304158492966, -2.324247325888213, -1.625453066846767, -0.926658807805319, -0.227864548763872, 0.470929710277574, 1.169723969319022, 1.868518228360469, 2.567312487401916, 3.266106746443364]
                            },
                            'bootstrap_sample_size': 20
                        }
                        }
            }
        ]
        for t in T:
            out = self.D.expect_column_bootstrapped_ks_test_p_value_to_be_greater_than(
                *t['args'], **t['kwargs'])
            if out['success'] != t['out']['success']:
                print("Test case error:")
                print(t)
                print(out)
            self.assertEqual(out['success'], t['out']['success'])
            if 'result_format' in t['kwargs'] and t['kwargs']['result_format'] == 'SUMMARY':
                self.assertTrue(np.allclose(
                    out['result']['details']['observed_cdf']['x'], t['out']['details']['observed_cdf']['x']))
                self.assertTrue(np.allclose(out['result']['details']['observed_cdf']
                                            ['cdf_values'], t['out']['details']['observed_cdf']['cdf_values']))
                self.assertTrue(np.allclose(
                    out['result']['details']['expected_cdf']['x'], t['out']['details']['expected_cdf']['x']))
                self.assertTrue(np.allclose(out['result']['details']['expected_cdf']
                                            ['cdf_values'], t['out']['details']['expected_cdf']['cdf_values']))
                self.assertTrue(np.allclose(
                    out['result']['details']['observed_partition']['bins'], t['out']['details']['observed_partition']['bins']))
                self.assertTrue(np.allclose(out['result']['details']['observed_partition']
                                            ['weights'], t['out']['details']['observed_partition']['weights']))
                self.assertTrue(np.allclose(
                    out['result']['details']['expected_partition']['bins'], t['out']['details']['expected_partition']['bins']))
                self.assertTrue(np.allclose(out['result']['details']['expected_partition']
                                            ['weights'], t['out']['details']['expected_partition']['weights']))

    def test_expect_column_bootstrapped_ks_test_p_value_to_be_greater_than_expanded_partitions(self):
        # Extend observed above and below expected
        out = self.D.expect_column_bootstrapped_ks_test_p_value_to_be_greater_than('norm_0_1', {'bins': np.linspace(-1, 1, 11), 'weights': [0.1] * 10},
                                                                                   result_format='SUMMARY')
        self.assertTrue(out['result']['details']['observed_cdf']['x'][0] < -1)
        self.assertTrue(out['result']['details']['observed_cdf']['x'][-1] > 1)
        # Extend observed below expected
        out = self.D.expect_column_bootstrapped_ks_test_p_value_to_be_greater_than('norm_0_1',
                                                                                   {'bins': np.linspace(-10, 1, 11), 'weights': [
                                                                                       0.1] * 10},
                                                                                   result_format='SUMMARY')
        self.assertTrue(out['result']['details']
                        ['observed_cdf']['x'][0] == -10)
        self.assertTrue(out['result']['details']['observed_cdf']['x'][-1] > 1)
        # Extend observed above expected
        out = self.D.expect_column_bootstrapped_ks_test_p_value_to_be_greater_than('norm_0_1',
                                                                                   {'bins': np.linspace(-1, 10, 11), 'weights': [
                                                                                       0.1] * 10},
                                                                                   result_format='SUMMARY')
        self.assertTrue(out['result']['details']['observed_cdf']['x'][0] < -1)
        self.assertTrue(out['result']['details']
                        ['observed_cdf']['x'][-1] == 10)
        # Extend expected above and below observed
        out = self.D.expect_column_bootstrapped_ks_test_p_value_to_be_greater_than('norm_0_1',
                                                                                   {'bins': np.linspace(-10, 10, 11), 'weights': [
                                                                                       0.1] * 10},
                                                                                   result_format='SUMMARY')
        self.assertTrue(out['result']['details']
                        ['observed_cdf']['x'][0] == -10)
        self.assertTrue(out['result']['details']
                        ['observed_cdf']['x'][-1] == 10)

    def test_expect_column_bootstrapped_ks_test_p_value_to_be_greater_than_bad_partition(self):
        with self.assertRaises(ValueError):
            self.D.expect_column_bootstrapped_ks_test_p_value_to_be_greater_than(
                'norm_0_1', {'bins': [-np.inf, 0, 1, 2, 3], 'weights': [0.25, 0.25, 0.25, 0.25]})

    def test_expect_column_kl_divergence_to_be_less_than_continuous_infinite_partition(self):
        # Manually build a partition extending to -Inf and Inf
        test_partition = self.test_partitions['norm_0_1_auto']
        test_partition['bins'] = [-np.inf] + test_partition['bins'] + [np.inf]
        scaled_weights = np.array(test_partition['weights']) * (1-0.01)
        test_partition['weights'] = [0.005] + scaled_weights.tolist() + [0.005]
        out = self.D.expect_column_kl_divergence_to_be_less_than(
            'norm_0_1', test_partition, 0.5, internal_weight_holdout=0.01)
        self.assertTrue(out['success'])

        # This should fail: tails have internal weight zero, which is highly unlikely.
        out = self.D.expect_column_kl_divergence_to_be_less_than(
            'norm_0_1', test_partition, 0.5)
        self.assertFalse(out['success'])

        # Build one-sided to infinity test partitions
        test_partition = {
            'bins': [-np.inf, 0, 1, 2, 3],
            'weights': [0.25, 0.25, 0.25, 0.25]
        }
        summary_expected_partition = {
            'bins': [0, 1, 2, 3],
            'weights': [0.25, 0.25, 0.25],
            'tail_weights':[0.25,0]
        }
        summary_observed_partition = {
            'bins': [0, 1, 2, 3],
            'weights': [0.25, 0.25, 0.25],
            'tail_weights':[0.25,0]
        }
        test_df = ge.dataset.PandasDataset(
            {'x': [-0.5, 0.5, 1.5, 2.5]})
        # This should succeed: our data match the partition
        out = test_df.expect_column_kl_divergence_to_be_less_than(
            'x', test_partition, 0.5, result_format='SUMMARY')
        self.assertTrue(out['success'])
        self.assertDictEqual(
            out['result']['details']['observed_partition'], summary_observed_partition)
        self.assertDictEqual(
            out['result']['details']['expected_partition'], summary_expected_partition)

        # Build one-sided to infinity test partitions
        test_partition = {
            'bins': [0, 1, 2, 3, np.inf],
            'weights': [0.25, 0.25, 0.25, 0.25]
        }
        summary_expected_partition = {
            'bins': [0, 1, 2, 3],
            'weights': [0.25, 0.25, 0.25],
            'tail_weights':[0,0.25]
        }
        summary_observed_partition = {
            'bins': [0, 1, 2, 3],
            'weights': [0.2, 0.2, 0.2],
            'tail_weights':[0.2,0.2]
        }
        test_df = ge.dataset.PandasDataset(
            {'x': [-0.5, 0.5, 1.5, 2.5, 3.5]})
        out = test_df.expect_column_kl_divergence_to_be_less_than(
            'x', test_partition, 0.5, result_format='SUMMARY')
        # This should fail: we expect zero weight less than 0
        self.assertFalse(out['success'])
        self.assertDictEqual(
            out['result']['details']['observed_partition'], summary_observed_partition)
        self.assertDictEqual(
            out['result']['details']['expected_partition'], summary_expected_partition)

        # Build two-sided to infinity test partition
        test_partition = {
            'bins': [-np.inf, 0, 1, 2, 3, np.inf],
            'weights': [0.1, 0.2, 0.4, 0.2, 0.1]
        }
        summary_expected_partition = {
            'bins': [ 0, 1, 2, 3],
            'weights': [0.2, 0.4, 0.2],
            'tail_weights':[0.1,0.1]
        }
        summary_observed_partition = {
            'bins': [0, 1, 2, 3],
            'weights': [0.2, 0.4, 0.2],
            'tail_weights':[0.1,0.1]
        }
        test_df = ge.dataset.PandasDataset(
            {'x': [-0.5, 0.5, 0.5, 1.5, 1.5, 1.5, 1.5, 2.5, 2.5, 3.5]})
        # This should succeed: our data match the partition
        out = test_df.expect_column_kl_divergence_to_be_less_than(
            'x', test_partition, 0.5, result_format='SUMMARY')
        self.assertTrue(out['success'])
        self.assertDictEqual(
            out['result']['details']['observed_partition'], summary_observed_partition)
        self.assertDictEqual(
            out['result']['details']['expected_partition'], summary_expected_partition)

        # Tail weight holdout is not defined for partitions already extending to infinity:
        with self.assertRaises(ValueError):
            test_df.expect_column_kl_divergence_to_be_less_than(
                'x', test_partition, 0.5, tail_weight_holdout=0.01)

    def test_expect_column_kl_divergence_to_be_less_than_continuous_serialized_infinite_partition(self):
        with open('./tests/test_sets/test_partition_serialized_infinity_bins.json', 'r') as infile:
            test_partition = json.loads(infile.read())['test_partition']

        summary_expected_partition = {
            'bins': [0, 1, 2, 3],
            'weights': [0.2, 0.4, 0.2],
            'tail_weights':[0.1,0.1]
        }
        summary_observed_partition = {
            'bins': [0, 1, 2, 3],
            'weights': [0.2, 0.4, 0.2],
            'tail_weights':[0.1,0.1]
        }
        test_df = ge.dataset.PandasDataset(
            {'x': [-0.5, 0.5, 0.5, 1.5, 1.5, 1.5, 1.5, 2.5, 2.5, 3.5]})
        # This should succeed: our data match the partition
        out = test_df.expect_column_kl_divergence_to_be_less_than(
            'x', test_partition, 0.5, result_format='SUMMARY')
        self.assertTrue(out['success'])
        self.assertDictEqual(
            out['result']['details']['observed_partition'], summary_observed_partition)
        self.assertDictEqual(
            out['result']['details']['expected_partition'], summary_expected_partition)

        # Confirm serialization of resulting expectations config
        expectation_config = test_df.get_expectations_config()
        found_expectation = False
        for expectation in expectation_config['expectations']:
            if 'expectation_type' in expectation and expectation['expectation_type'] == 'expect_column_kl_divergence_to_be_less_than':
                self.assertEqual(
                    json.dumps(expectation['kwargs']
                               ['partition_object']['bins']),
                    '[-Infinity, 0, 1, 2, 3, Infinity]'
                )
                found_expectation = True
        self.assertTrue(found_expectation)

    def test_expect_column_kl_divergence_to_be_less_than_continuous(self):
        T = [
                {
                    'args': ['norm_0_1'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_auto'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':True, 'observed_value': 'NOTTESTED'}
                },
                {
                    'args': ['norm_0_1'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_uniform'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':True, 'observed_value': 'NOTTESTED'}
                },
                {
                    'args': ['norm_0_1'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_ntile'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':True, 'observed_value': 'NOTTESTED'}
                },
                ## Note higher threshold example for kde
                {
                    'args': ['norm_0_1'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_kde'],
                              "threshold": 0.3,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':True, 'observed_value': 'NOTTESTED'}
                },
                {
                    'args': ['norm_1_1'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_auto'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':False, 'observed_value': 'NOTTESTED'}
                },
                {
                    'args': ['norm_1_1'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_uniform'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 1e-5,
                              "internal_weight_holdout": 1e-5},
                    'out':{'success':False, 'observed_value': 'NOTTESTED'}
                },
                {
                    'args': ['norm_1_1'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_ntile'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':False, 'observed_value': 'NOTTESTED'}
                },
                {
                    'args': ['norm_1_1'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_kde'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':False, 'observed_value': 'NOTTESTED'}
                },
                {
                    'args': ['bimodal'],
                    'kwargs':{"partition_object": self.test_partitions['bimodal_auto'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':True, 'observed_value': 'NOTTESTED'}
                },
                {
                    'args': ['bimodal'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_auto'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':False, 'observed_value': "NOTTESTED"}
                },
                {
                    'args': ['bimodal'],
                    'kwargs':{"partition_object": self.test_partitions['norm_0_1_uniform'],
                              "threshold": 0.1,
                              "tail_weight_holdout": 0.01,
                              "internal_weight_holdout": 0.01},
                    'out':{'success':False, 'observed_value': "NOTTESTED"}
                },
                {
                    'args': ['bimodal'],
                    'kwargs': {"partition_object": self.test_partitions['norm_0_1_uniform'],
                               "threshold": 0.1,
                               "tail_weight_holdout": 0.01,
                               "internal_weight_holdout": 0.01,
                               "result_format": "SUMMARY"},
                    'out': {'success': False, 'observed_value': "NOTTESTED",
                            'details':
                                {'observed_partition':
                                     {'weights': [0.001, 0.006, 0.022, 0.07, 0.107, 0.146, 0.098, 0.04, 0.01, 0.0],
                                      'bins': [-3.721835843971108, -3.02304158492966, -2.324247325888213, -1.625453066846767, -0.926658807805319, -0.227864548763872, 0.470929710277574, 1.169723969319022, 1.868518228360469, 2.567312487401916, 3.266106746443364],
                                      'tail_weights':[0.0,0.5]
                                     },
                                 'missing_percent': 0.0,
                                 'element_count': 1000,
                                 'missing_count': 0,
                                 'expected_partition': {'bins': [-3.721835843971108, -3.02304158492966, -2.324247325888213, -1.625453066846767, -0.926658807805319, -0.227864548763872, 0.470929710277574, 1.169723969319022, 1.868518228360469, 2.567312487401916, 3.266106746443364],
                                                        'weights': [0.00098, 0.00784, 0.04606, 0.12544, 0.24009999999999998, 0.24402, 0.20776, 0.07742, 0.02352, 0.00686],
                                                        'tail_weights':[0.005,0.005]
                                                        }
                                 }
                    }
                }
        ]
        for t in T:
            out = self.D.expect_column_kl_divergence_to_be_less_than(
                *t['args'], **t['kwargs'])
            if t['out']['observed_value'] != 'NOTTESTED':
                if not np.allclose(out['observed_value'], t['out']['observed_value']):
                    print("Test case error:")
                    print(t)
                    print(out)
                self.assertTrue(np.allclose(
                    out['observed_value'], t['out']['observed_value']))
            if 'result_format' in t['kwargs'] and t['kwargs']['result_format'] == 'SUMMARY':
                self.assertTrue(np.allclose(out['result']['details']['observed_partition']['bins'],t['out']['details']['observed_partition']['bins']))
                self.assertTrue(np.allclose(out['result']['details']['observed_partition']['weights'],t['out']['details']['observed_partition']['weights']))
                self.assertTrue(np.allclose(out['result']['details']['observed_partition']['tail_weights'],t['out']['details']['observed_partition']['tail_weights']))
                self.assertTrue(np.allclose(out['result']['details']['expected_partition']['bins'],t['out']['details']['expected_partition']['bins']))
                self.assertTrue(np.allclose(out['result']['details']['expected_partition']['weights'],t['out']['details']['expected_partition']['weights']))
                self.assertTrue(np.allclose(out['result']['details']['expected_partition']['tail_weights'],t['out']['details']['expected_partition']['tail_weights']))

            if not out['success'] == t['out']['success']:
                print("Test case error:")
                print(t)
                print(out)
            self.assertEqual(out['success'], t['out']['success'])

    def test_expect_column_kl_divergence_to_be_less_than_bad_parameters(self):
        with self.assertRaises(ValueError):
            self.D.expect_column_kl_divergence_to_be_less_than(
                'norm_0_1', {}, threshold=0.1)
        with self.assertRaises(ValueError):
            self.D.expect_column_kl_divergence_to_be_less_than(
                'norm_0_1', self.test_partitions['norm_0_1_auto'])
        with self.assertRaises(ValueError):
            self.D.expect_column_kl_divergence_to_be_less_than(
                'norm_0_1', self.test_partitions['norm_0_1_auto'], threshold=0.1, tail_weight_holdout=2)
        with self.assertRaises(ValueError):
            self.D.expect_column_kl_divergence_to_be_less_than(
                'norm_0_1', self.test_partitions['norm_0_1_auto'], threshold=0.1, internal_weight_holdout=2)
        with self.assertRaises(ValueError):
            self.D.expect_column_kl_divergence_to_be_less_than(
                'categorical_fixed', self.test_partitions['categorical_fixed'], threshold=0.1, internal_weight_holdout=0.01)

    def test_expect_column_kl_divergence_to_be_less_than_infinite_observed_value(self):

        # Test with discrete distribution
        discrete_df = ge.dataset.PandasDataset(
            {"x1": ['a', 'a', 'a', 'a', 'b', 'b', 'b', 'b', 'c', 'c']})
        discrete_partition_object = {"weights": [0.5, 0.5],
                                     "values": ["a", "b"]}
        discrete_out = discrete_df.expect_column_kl_divergence_to_be_less_than("x1",
                                                                               discrete_partition_object,
                                                                               0.05)
        discrete_out_json = json.dumps(
            discrete_out['result']['observed_value'])
        self.assertEqual(discrete_out['result']['observed_value'], None)
        self.assertEqual(discrete_out_json,
                         'null')

        # Test with continuous distribution
        continuous_df = ge.dataset.PandasDataset({"x2": [-1.5, -1.5, -1.35, -1.2, -1.19, -1.12, -1.04, -1, 0, 0, 0, 0.25,
                                                         0.34, 1, 1.1, 1.13, 1.22, 1.34, 1.4, 1.46, 1.5, 1.5, 1.5]})

        continuous_partition_object1 = {"weights": [0.1, 0.4, 0, 0, 0.4, 0.1],
                                        "bins": [-2, -1.5, -1, 0, 1, 1.5, 2]}  # Partition sets weight to zero for to a non-empty bin

        continuous_out1 = continuous_df.expect_column_kl_divergence_to_be_less_than("x2",
                                                                                    continuous_partition_object1,
                                                                                    0.05)
        continuous_out_json1 = json.dumps(
            continuous_out1['result']['observed_value'])

        self.assertEqual(continuous_out1['result']['observed_value'], None)
        self.assertEqual(continuous_out_json1,
                         'null')

        continuous_partition_object2 = {"weights": [0.25, 0.25, 0.25, 0.25],
                                        "bins": [-1, 0, 1, 1.5, 2]}  # Partition bins do not cover all of data set

        continuous_out2 = continuous_df.expect_column_kl_divergence_to_be_less_than("x2",
                                                                                    continuous_partition_object2,
                                                                                    0.05)
        continuous_out_json2 = json.dumps(
            continuous_out2['result']['observed_value'])

        self.assertEqual(continuous_out2['result']['observed_value'], None)
        self.assertEqual(continuous_out_json2,
                         'null')
        
        
    def test_expect_column_kl_divergence_to_be_less_than_infinite_return_bins(self):
    
    
        continuous_df=ge.dataset.PandasDataset({"x":[-1.95, 1.03, 1.00, 0.81, -2.27,  0.52, 2.45, -1.19,
                                                     -0.17, -1.54, 2.20, -2.66,  1.71,  1.59, 2.19]})
        
        
        #Both endpoints at infinity, no tail_holdout or internal_holdout
        continuous_partition_object={"weights":[0.3,0.15,0.05,0.05,0.2,0.25],
            "bins":[-np.inf,-2,-1,0,1,2,np.inf]}
        
        
        expected_details={
                            "observed_partition": {
                                # return expected_bins, since we used those bins to compute the observed_weights
                                "bins": [-2,-1,0,1,2],
                                "weights": [3/15.0, 1/15.0, 2/15.0, 4/15.0],
                                "tail_weights": [2/15.0,3/15.0]
                            },
                            "expected_partition": {
                                "bins": [-2,-1,0,1,2],
                                "weights": [0.15,0.05,0.05,0.2],
                                "tail_weights":[0.3,0.25]
                            }
                        }
                            
        actual_details=continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                                 continuous_partition_object,
                                                                                 0.05,
                                                                                 tail_weight_holdout=0, 
                                                                                 internal_weight_holdout=0,
                                                                                 result_format="COMPLETE")["result"]["details"]
        assertDeepAlmostEqual(expected_details,actual_details)
        self.assertTrue(is_valid_continuous_partition_object(actual_details["observed_partition"]))
        self.assertTrue(is_valid_continuous_partition_object(actual_details["expected_partition"]))
        
        
        
        #Both endpoints bounded, no tail_holdout or internal_holdout
        
        continuous_partition_object={"weights":[0.3,0.15,0.05,0.05,0.2,0.25],
            "bins":[-3,-2,-1,0,1,2,3]}
        
        
        expected_details={
                            "observed_partition": {
                                # return expected_bins, since we used those bins to compute the observed_weights
                                "bins": [-3,-2,-1,0,1,2,3],
                                "weights": [2/15.0,3/15.0, 1/15.0, 2/15.0, 4/15.0,3/15.0],
                                "tail_weights": [0,0]
                            },
                            "expected_partition": {
                                "bins": [-3,-2,-1,0,1,2,3],
                                "weights": [0.3,0.15,0.05,0.05,0.2,0.25],
                                "tail_weights":[0,0]
                            }
                        }
                            
        actual_details=continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                                 continuous_partition_object,
                                                                                 0.05,
                                                                                 tail_weight_holdout=0, 
                                                                                 internal_weight_holdout=0,
                                                                                 result_format="COMPLETE")["result"]["details"]
        assertDeepAlmostEqual(expected_details,actual_details)
        self.assertTrue(is_valid_continuous_partition_object(actual_details["observed_partition"]))
        self.assertTrue(is_valid_continuous_partition_object(actual_details["expected_partition"]))
        
        
        #Infinite KL Divergence, no tail_holdout or internal_holdout
        
        continuous_partition_object={"weights":[0.45,0.05,0.05,0.2,0.25],
            "bins":[-2,-1,0,1,2,3]}
        
        
        expected_details={
                            "observed_partition": {
                                # return expected_bins, since we used those bins to compute the observed_weights
                                "bins": [-2,-1,0,1,2,3],
                                "weights": [3/15.0, 1/15.0, 2/15.0, 4/15.0,3/15.0],
                                "tail_weights": [2/15.0,0]
                            },
                            "expected_partition": {
                                "bins": [-2,-1,0,1,2,3],
                                "weights": [0.45,0.05,0.05,0.2,0.25],
                                "tail_weights":[0,0]
                            }
                        }
                            
        actual_details=continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                                 continuous_partition_object,
                                                                                 0.05,
                                                                                 tail_weight_holdout=0, 
                                                                                 internal_weight_holdout=0,
                                                                                 result_format="COMPLETE")["result"]["details"]
        assertDeepAlmostEqual(expected_details,actual_details)
        self.assertTrue(is_valid_continuous_partition_object(actual_details["observed_partition"]))
        self.assertTrue(is_valid_continuous_partition_object(actual_details["expected_partition"]))
        
        
    
        #Both endpoints at infinity, non-zero tail hold_out
        continuous_partition_object={"weights":[0.3,0.15,0.05,0.05,0.2,0.25],
            "bins":[-np.inf,-2,-1,0,1,2,np.inf]}
        
        with self.assertRaises(ValueError):
            
            actual_details=continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                                 continuous_partition_object,
                                                                                 0.05,
                                                                                 tail_weight_holdout=0.01, 
                                                                                 internal_weight_holdout=0,
                                                                                 result_format="COMPLETE")["result"]["details"]
            
            
        
        #Upper at infinity, non-zero tail hold_out
        continuous_partition_object={"weights":[0.3,0.15,0.05,0.05,0.2,0.25],
            "bins":[-3,-2,-1,0,1,2,np.inf]}
        
        
        expected_details={
                            "observed_partition": {
                                # return expected_bins, since we used those bins to compute the observed_weights
                                "bins": [-3,-2,-1,0,1,2],
                                "weights": [2/15.0,3/15.0, 1/15.0, 2/15.0, 4/15.0],
                                "tail_weights": [0,3/15.0]
                            },
                            "expected_partition": {
                                "bins": [-3,-2,-1,0,1,2],
                                "weights": [0.3*0.98,0.15*0.98,0.05*0.98,0.05*0.98,0.2*0.98],
                                "tail_weights":[0.02,0.98*0.25]
                            }
                        }
                            
        actual_details=continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                                 continuous_partition_object,
                                                                                 0.05,
                                                                                 tail_weight_holdout=0.02, 
                                                                                 internal_weight_holdout=0,
                                                                                 result_format="COMPLETE")["result"]["details"]
        assertDeepAlmostEqual(expected_details,actual_details)
        
        self.assertTrue(is_valid_continuous_partition_object(actual_details["observed_partition"]))
        self.assertTrue(is_valid_continuous_partition_object(actual_details["expected_partition"]))
        
    
            
        
        #Lower at infinity, non-zero tail holdout
        continuous_partition_object={"weights":[0.3,0.15,0.05,0.05,0.45],
            "bins":[-np.inf,-2,-1,0,1,2]}
        
        
        expected_details={
                            "observed_partition": {
                                # return expected_bins, since we used those bins to compute the observed_weights
                                "bins": [-2,-1,0,1,2],
                                "weights": [3/15.0, 1/15.0, 2/15.0, 4/15.0],
                                "tail_weights": [2/15.0,3/15.0]
                            },
                            "expected_partition": {
                                "bins": [-2,-1,0,1,2],
                                "weights": [0.15*0.95,0.05*0.95,0.05*0.95,0.45*0.95],
                                "tail_weights":[0.3*0.95,0.05]
                            }
                        }
                            
        actual_details=continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                                 continuous_partition_object,
                                                                                 0.05,
                                                                                 tail_weight_holdout=0.05, 
                                                                                 internal_weight_holdout=0,
                                                                                 result_format="COMPLETE")["result"]["details"]
        assertDeepAlmostEqual(expected_details,actual_details)
        
        self.assertTrue(is_valid_continuous_partition_object(actual_details["observed_partition"]))
        self.assertTrue(is_valid_continuous_partition_object(actual_details["expected_partition"]))
        
        
        #Bounded Endpoints, non-zero tail holdout, non-zero internal holdout
        continuous_partition_object={"weights":[0.3,0.15,0.0,0.10,0.45],
            "bins":[-3,-2,-1,0,1,2]}
        
        
        expected_details= {
                            "observed_partition": {
                                # return expected_bins, since we used those bins to compute the observed_weights
                                "bins": [-3,-2,-1,0,1,2],
                                "weights": [2/15.0,3/15.0, 1/15.0, 2/15.0, 4/15.0],
                                "tail_weights": [0.0,3/15.0]
                            },
                            "expected_partition": {
                                "bins": [-3,-2,-1,0,1,2],
                                "weights": [0.3*0.9,0.15*0.9,0.05,0.10*0.9,0.45*0.9],
                                "tail_weights":[0.05/2,0.05/2]
                            }
                        }
                            
        actual_details=continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                                 continuous_partition_object,
                                                                                 0.05,
                                                                                 tail_weight_holdout=0.05, 
                                                                                 internal_weight_holdout=0.05,
                                                                                 result_format="COMPLETE")["result"]["details"]
        
        assertDeepAlmostEqual(expected_details,actual_details)
        
        self.assertTrue(is_valid_continuous_partition_object(actual_details["observed_partition"]))
        self.assertTrue(is_valid_continuous_partition_object(actual_details["expected_partition"]))
        
        
    def test_expect_column_kl_divergence_to_be_less_than_user_tail_weights(self):
    
        continuous_df=ge.dataset.PandasDataset({"x":[-1.95, 1.03, 1.00, 0.81, -2.27,  0.52, 2.45, -1.19,
                                                 -0.17, -1.54, 2.20, -2.66,  1.71,  1.59, 2.19]})
    
    
    
    #This should exit without error 
    
        continuous_partition_object={"weights":[0.3,0.15,0.0,0.10,0.15],
        "bins":[-3,-2,-1,0,1,2], "tail_weights":[0.15,0.15]}
        
        continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                  continuous_partition_object,
                                                                  0.05)
        
            
    #Error: Only one tail weight
        continuous_partition_object={"weights":[0.3,0.15,0.0,0.10,0.30],
                                     "bins":[-3,-2,-1,0,1,2], "tail_weights":[0.15]}
    
        self.assertFalse(is_valid_continuous_partition_object(continuous_partition_object))
                                                                  

        #Error: Use of tail_weights with infinite end points partition
        continuous_partition_object={"weights":[0.3,0.15,0.0,0.10,0.15],
        "bins":[-3,-2,-1,0,1,np.inf], "tail_weights":[0.15,0.15]}
        
        with self.assertRaises(ValueError):
            continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                      continuous_partition_object,
                                                                      0.05) 
        
        
        #Error: Use of tail_weights and tail_weight_holdout
        
        continuous_partition_object={"weights":[0.3,0.15,0.0,0.10,0.15],
        "bins":[-3,-2,-1,0,1,2], "tail_weights":[0.15,0.15]}
        
        with self.assertRaises(ValueError):
            continuous_df.expect_column_kl_divergence_to_be_less_than("x",
                                                                      continuous_partition_object,
                                                                      0.05,
                                                                      tail_weight_holdout=0.01)
        
        #Error: Tail_weights and weights don't add to one
        
        continuous_partition_object={"weights":[0.3,0.15,0.0,0.10,0.16],
        "bins":[-3,-2,-1,0,1,2], "tail_weights":[0.15,0.15]}
        
        self.assertFalse(is_valid_continuous_partition_object(continuous_partition_object))
        
        
        

    def test_expect_column_bootstrapped_ks_test_p_value_to_be_greater_than_bad_parameters(self):
        with self.assertRaises(ValueError):
            self.D.expect_column_bootstrapped_ks_test_p_value_to_be_greater_than(
                'norm_0_1', self.test_partitions['categorical_fixed'])
        test_partition = ge.dataset.util.kde_partition_data(
            self.D['norm_0_1'], estimate_tails=False)
        with self.assertRaises(ValueError):
            self.D.expect_column_bootstrapped_ks_test_p_value_to_be_greater_than(
                'norm_0_1', test_partition)

    def test_expect_column_chisquare_test_p_value_to_be_greater_than_bad_parameters(self):
        with self.assertRaises(ValueError):
            self.D.expect_column_chisquare_test_p_value_to_be_greater_than(
                'categorical_fixed', self.test_partitions['norm_0_1_auto'])


if __name__ == "__main__":
    unittest.main()
